/*
 * VegaFusion
 * Copyright (C) 2022 VegaFusion Technologies LLC
 *
 * This program is distributed under multiple licenses.
 * Please consult the license documentation provided alongside
 * this program the details of the active license.
 */
/*!
## Array Functions
Functions for working with arrays of values.

See https://vega.github.io/vega/docs/expressions/#array-functions
 */
pub mod length;
pub mod span;
