/*
 * VegaFusion
 * Copyright (C) 2022 VegaFusion Technologies LLC
 *
 * This program is distributed under multiple licenses.
 * Please consult the license documentation provided alongside
 * this program the details of the active license.
 */
/*!
## Math Functions
Basic mathematical functions.

See https://vega.github.io/vega/docs/expressions/#math-functions
 */
pub mod isfinite;
pub mod isnan;
pub mod pow;
